#ifndef KACCESSDSO_H
#define KACCESSDSO_H
#include "KMark.h"
#include "KAcl.h"
#include "ksapi.h"
#include "KDsoExtend.h"
#include "KAccessDsoSupport.h"
#include "KBufferFetchObject.h"

#ifdef ENABLE_KSAPI_FILTER

class KAccessDso
{
public:
	KAccessDso(kgl_access *access, KDsoExtend *dso,int notify_type)
	{
		this->access = access;
		this->dso = dso;
		init_access_dso_support(&ctx,notify_type);
		ctx.module = access->create_ctx();
		this->notify_type = notify_type;
	}
	~KAccessDso()
	{
		if  (ctx.module)  {
			access->free_ctx(ctx.module);
		}
	}
	int32_t Shutdown();
	void Init();
	KF_STATUS_TYPE process(KHttpRequest *rq, KAccessContext *cn,DWORD notify);
	KAccessDso *newInstance()
	{
		return new KAccessDso(access,dso,notify_type);
	}
	const char *getName()
	{
		return access->name;
	}
	std::string getHtml(KModel *model);
	std::string getDisplay();
	void editHtml(std::map<std::string, std::string> &attribute,bool html);
	bool startCharacter(KXmlContext *context, char *character, int len);
	void buildXML(std::stringstream &s);
	bool endElement(KXmlContext *context);
	friend class KAccessDsoMark;
	friend class KAccessDsoAcl;
	KDsoExtend *dso;
private:
	std::string build(KF_ACCESS_BUILD_TYPE type);
	kgl_access *access;
	kgl_access_context ctx;
	int notify_type;
};
class KAccessDsoMark : public KMark
{
public:
	KAccessDsoMark(KAccessDso *ad)
	{
		this->ad = ad;
	}
	~KAccessDsoMark()
	{
		delete ad;
	}
	int32_t Shutdown()
	{
		return ad->Shutdown();
	}
	void Init()
	{
		ad->Init();
	}
	bool mark(KHttpRequest *rq, KHttpObject *obj,const int chainJumpType, int &jumpType)
	{
		KAccessContext cn;
		switch (ad->process(rq, &cn,obj? KF_NOTIFY_RESPONSE_MARK:KF_NOTIFY_REQUEST_MARK)) {
		case KF_STATUS_REQ_ACCESS_TRUE:
			return true;
		case KF_STATUS_REQ_ACCESS_FALSE:
			return false;
		case KF_STATUS_REQ_FINISHED:
			SET(rq->flags, RQ_CONNECTION_CLOSE);
		default:
			if (cn.buffer) {
				kassert(TEST(ad->notify_type, KF_NOTIFY_RESPONSE_MARK | KF_NOTIFY_RESPONSE_ACL ) == 0);
				rq->closeFetchObject();				
				rq->responseConnection();
				rq->responseHeader(kgl_expand_string("Content-Length"), cn.buffer->getLen());
				rq->AppendFetchObject(new KBufferFetchObject(cn.buffer));
				if (rq->status_code == 0) {
					rq->responseStatus(STATUS_OK);
				}
			}
			jumpType = JUMP_FINISHED;
			return true;
		}
	}
	KMark *newInstance()
	{
		return new KAccessDsoMark(ad->newInstance());
	}
	const char *getName()
	{
		return ad->getName();
	}
	std::string getHtml(KModel *model)
	{
		return ad->getHtml(model);
	}
	std::string getDisplay()
	{
		return ad->getDisplay();
	}
	void editHtml(std::map<std::string, std::string> &attribute,bool html);
	void buildXML(std::stringstream &s)
	{
		ad->buildXML(s);
	}
	bool startCharacter(KXmlContext *context, char *character, int len)
	{
		return ad->startCharacter(context, character, len);
	}
	bool endElement(KXmlContext *context)
	{
		return ad->endElement(context);
	}
private:
	KAccessDso *ad;
};
class KAccessDsoAcl : public KAcl
{
public:
	KAccessDsoAcl(KAccessDso *ad)
	{
		this->ad = ad;
	}
	~KAccessDsoAcl()
	{
		delete ad;
	}
	bool match(KHttpRequest *rq, KHttpObject *obj)
	{
		KAccessContext cn;
		KF_STATUS_TYPE ret = ad->process(rq, &cn, obj ? KF_NOTIFY_RESPONSE_ACL : KF_NOTIFY_REQUEST_ACL);
		switch (ret) {
		case KF_STATUS_REQ_ACCESS_FALSE:
			return false;
		case KF_STATUS_REQ_ACCESS_TRUE:
			return true;
		default:
			klog(KLOG_ERR, "access dso [%s] process result=[%d] is illegal.\n", this->getName(), ret);
		}
		return false;
	}
	KAcl *newInstance()
	{
		return new KAccessDsoAcl(ad->newInstance());
	}
	const char *getName()
	{
		return ad->getName();
	}
	std::string getHtml(KModel *model)
	{
		return ad->getHtml(model);
	}
	std::string getDisplay()
	{
		return ad->getDisplay();
	}
	void editHtml(std::map<std::string, std::string> &attribute,bool html)
	{
		ad->editHtml(attribute,html);
	}
	void buildXML(std::stringstream &s)
	{
		ad->buildXML(s);
	}
	bool startCharacter(KXmlContext *context, char *character, int len)
	{
		return ad->startCharacter(context, character, len);
	}
	bool endElement(KXmlContext *context)
	{
		return ad->endElement(context);
	}
private:
	KAccessDso *ad;
};
#endif
#endif
