#include <errno.h>
#include "KAsyncFetchObject.h"
#include "http.h"
#include "kselector.h"
#include "KAsyncWorker.h"
#ifdef _WIN32
#include "KIOCPSelector.h"
#endif
#include "KSingleAcserver.h"
#include "KCdnContainer.h"
#include "KSimulateRequest.h"
#include "KHttpProxyFetchObject.h"
#include "KHttp2.h"

#include "KHttpFieldValue.h"

#ifdef ENABLE_UPSTREAM_SSL
kev_result result_ssl_connect(KOPAQUE data, void *arg, int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	if (got < 0) {
		return fo->handleUpstreamError(rq, STATUS_BAD_GATEWAY, "cann't connect to host use ssl", got);
	}
	assert(fo->client);

	return fo->sendHead(rq);
}
#endif
kev_result next_connection_result(KOPAQUE data, void *arg, int got)
{
	KHttpRequest *rq = (KHttpRequest *)data;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)arg;
	return fo->connect_result(rq, got!=0);
}
#ifdef ENABLE_PROXY_PROTOCOL
kev_result proxyTcpConnect(KHttpRequest *rq,KAsyncFetchObject *fo)
{
	char ips[MAXIPLEN];
	kconnection *cn = rq->sink->GetConnection();
	if (cn->proxy == NULL || cn->proxy->dst == NULL) {
		return	stageEndRequest(rq);	
	}
	if (!ksocket_sockaddr_ip(cn->proxy->dst, ips, MAXIPLEN-1)) {
		return stageEndRequest(rq);
	}
	uint16_t port = ntohs(cn->proxy->dst->v4.sin_port);
	KSingleAcserver *server = new KSingleAcserver;
	server->proto = Proto_tcp;
	server->sockHelper->setHostPort(ips, port, NULL);
	server->sockHelper->setLifeTime(0);
	return server->connect(rq,fo);
}
#endif
kev_result proxy_connect(KHttpRequest *rq,KAsyncFetchObject *fo)
{
#ifdef ENABLE_PROXY_PROTOCOL
	if (TEST(rq->GetWorkModel(), WORK_MODEL_PROXY|WORK_MODEL_SSL_PROXY)) {
		return proxyTcpConnect(rq,fo);
	}
#endif
	//KAsyncFetchObject *fo = static_cast<KAsyncFetchObject *>(rq->fetchObj);
	const char *ip = rq->bind_ip;
	KUrl *url = (TEST(rq->filter_flags,RF_PROXY_RAW_URL)?&rq->raw_url:rq->url);
	const char *host = url->host;
	u_short port = url->port;
	const char *ssl = NULL;
	int life_time = 2;
#ifdef IP_TRANSPARENT
#ifdef ENABLE_TPROXY
	char mip[MAXIPLEN];
	if (TEST(rq->GetWorkModel(),WORK_MODEL_TPROXY) && TEST(rq->filter_flags,RF_TPROXY_TRUST_DNS)) {
		if (TEST(rq->filter_flags,RF_TPROXY_UPSTREAM)) {
			if (ip==NULL) {
				ip = rq->getClientIp();
			}
		}
		sockaddr_i s_sockaddr;
		socklen_t addr_len = sizeof(sockaddr_i);
		::getsockname(rq->sink->GetConnection()->st.fd, (struct sockaddr *) &s_sockaddr, &addr_len);
		ksocket_sockaddr_ip(&s_sockaddr, mip, MAXIPLEN);
		host = mip;
#ifdef KSOCKET_IPV6
		if (s_sockaddr.v4.sin_family == PF_INET6) {
			port = ntohs(s_sockaddr.v6.sin6_port);
		} else
#endif
		port = ntohs(s_sockaddr.v4.sin_port);
	}
#endif
#endif
	if (TEST(url->flags, KGL_URL_ORIG_SSL)) {
		ssl = "s";
	}
#ifdef ENABLE_SIMULATE_HTTP
	/* simuate request must replace host and port */
	if (rq->ctx->simulate) {
		KSimulateSink *ss = static_cast<KSimulateSink *>(rq->sink);
		if (ss->host && *ss->host) {
			host = ss->host;
			if (ss->port>0) {
				port = ss->port;
			}
			life_time = ss->life_time;
		}
	}
#endif
	KRedirect *sa = server_container->refsRedirect(ip,host,port,ssl,life_time,Proto_http);
	if (sa==NULL) {
		return fo->connectCallBack(rq,NULL);
	}
	kev_result ret = sa->connect(rq,fo);
	sa->release();
	return ret;	
}
kev_result resultUpStreamReadBodyResult(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->handleReadBody(rq,got);
}
kev_result resultUpstreamSendPost(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->handleSendPost(rq,got);
}
int bufferUpstreamSendPost(KOPAQUE data, void *arg,LPWSABUF buf,int bufCount)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->getPostRBuffer(rq,buf,bufCount);
}
kev_result resultUpstreamSendHead(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->handleSendHead(rq,got);
}
kev_result request_connection_broken(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)data;
	rq->ctx->read_huped = true;
	KFetchObject *fo = rq->GetFetchObject();
	if (fo==NULL) {
		return kev_err;
	}
	fo->Shutdown(rq);
	return kev_err;
}
int bufferUpstreamSendHead(KOPAQUE data, void *arg,LPWSABUF buf,int bufCount)
{
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->buffer->getReadBuffer(buf,bufCount);
}
kev_result resultUpstreamReadHead(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->handleReadHead(rq,got);
}
int bufferUpstreamRead(KOPAQUE data, void *arg,LPWSABUF buf,int bufCount)
{
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	int len;
	buf[0].iov_base = (char *)fo->getUpstreamBuffer(&len);
	buf[0].iov_len = len;
	return 1;
}
kev_result resultUpstreamConnectResult(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->handleConnectResult(rq,got);
}
kev_result resultUpstreamHttp2ReadHeader(KOPAQUE data, void *arg,int got)
{
	KHttpRequest *rq = (KHttpRequest *)arg;
	KAsyncFetchObject *fo = (KAsyncFetchObject *)data;
	return fo->handleHttp2ReadHead(rq,got);
}
kev_result KAsyncFetchObject::handleConnectResult(KHttpRequest *rq,int got)
{
	if (got<0) {
		return handleUpstreamError(rq,STATUS_GATEWAY_TIMEOUT,"connect to remote host time out",got);
	}
#ifdef ENABLE_UPSTREAM_SSL
	kconnection *c = client->GetConnection();
	if (kconnection_is_ssl(c)) {
		return kconnection_ssl_handshake(c, result_ssl_connect, rq);
	}
#endif
	return sendHead(rq);
}
void KAsyncFetchObject::ResetBuffer()
{
	if (buffer) {
		delete buffer;
		buffer = NULL;
	}
	if (us_buffer.buf) {
		xfree(us_buffer.buf);
	}
	memset(&us_buffer, 0, sizeof(us_buffer));
	memset(&parser, 0, sizeof(parser));
	memset(&pop_header, 0, sizeof(pop_header));
}
kev_result KAsyncFetchObject::retryOpen(KHttpRequest *rq)
{
	if (client) {
		client->Gc(-1,0);
		client = NULL;
	}
	ResetBuffer();
	return InternalOpen(rq);
}
kev_result KAsyncFetchObject::InternalOpen(KHttpRequest *rq)
{
	pop_header.keep_alive_time_out = 0;
	rq->ctx->upstream_connection_keep_alive = 1;
	if (brd == NULL) {
		return proxy_connect(rq, this);
	}
	if (!brd->rd->enable) {
		return gate->f->push_message(gate, rq, KGL_MSG_ERROR, STATUS_SERVICE_UNAVAILABLE, "extend is disable");
	}
	badStage = BadStage_Connect;
	return brd->rd->connect(rq, this);
}
kev_result KAsyncFetchObject::Open(KHttpRequest *rq, kgl_push_gate *gate)
{
	KFetchObject::Open(rq,gate);
	return InternalOpen(rq);
}
void KAsyncFetchObject::read_hup(KHttpRequest *rq)
{
	if (!conf.read_hup) {
		return;
	}
	rq->sink->ReadHup(this, request_connection_broken);
}
kev_result KAsyncFetchObject::connect_result(KHttpRequest *rq, bool half_connection)
{
	kassert(kselector_is_same_thread(rq->sink->GetSelector()));
	if (this->client == NULL) {
		return gate->f->push_message(gate, rq, KGL_MSG_ERROR, STATUS_GATEWAY_TIMEOUT, "Cann't connect to remote host");
	}
	client->BindOpaque(this);
	client->SetTimeOut(rq->sink->GetTimeOut());
	client->BindSelector(rq->sink->GetSelector());
	read_hup(rq);
	if (half_connection) {
		return client->Connect(rq,resultUpstreamConnectResult);
	}
	return sendHead(rq);
}
kev_result KAsyncFetchObject::connectCallBack(KHttpRequest *rq,KUpstream *client,bool half_connection)
{
	assert(this->client == NULL);
	this->client = client;
	kselector *selector = rq->sink->GetSelector();
	if (kselector_is_same_thread(selector)) {
		return connect_result(rq, half_connection);
	}
	kgl_selector_module.next(selector, rq, next_connection_result, this, half_connection);
	return kev_ok;
}
kev_result KAsyncFetchObject::sendHead(KHttpRequest *rq)
{
	
	badStage = BadStage_TrySend;
	if (gate->f->get_post_left(gate, rq) == -1 && !client->IsMultiStream()) {
		chunk_post = 1;
	}
#if 0
	if (rq->left_read != 0 && !client->IsMultiStream()) {
		if (TEST(rq->flags, RQ_INPUT_CHUNKED)) {
			rq->sink->SetChunkRawRead();
		} else if (rq->left_read == -1) {
			chunk_post = 1;
		}
	}
#endif
	if (buffer==NULL) {
		buildHead(rq);
		kassert(buffer);
		
	}
	if (buffer->startRead()==0) {
		return sendHeadSuccess(rq);
	}
	client->SetDelay();
	return client->Write(rq,resultUpstreamSendHead,bufferUpstreamSendHead);
}
kev_result KAsyncFetchObject::continueReadBody(KHttpRequest *rq)
{
	if (rq->ctx->connection_upgrade) {
		return client->Read(rq, resultUpStreamReadBodyResult, bufferUpstreamRead);
	}
	if (!checkContinueReadBody(rq)) {
		return gate->f->push_body_finish(gate, rq, KGL_OK);
	}
	read_hup(rq);
	return client->Read(rq,resultUpStreamReadBodyResult, bufferUpstreamRead);
}
kev_result KAsyncFetchObject::Read(KHttpRequest *rq)
{	
	if (us_buffer.used > 0) {
		return ParseBody(rq);
	}
	return continueReadBody(rq);
}
kev_result KAsyncFetchObject::handleReadBody(KHttpRequest *rq, int got)
{
	if (got <= 0) {
		if (rq->ctx->connection_upgrade) {
			return shutdown(rq);
		}
		pop_header.keep_alive_time_out = -1;
		return gate->f->push_body_finish(gate, rq, got==0? KGL_OK: KGL_ESOCKET_BROKEN);
	}
	ks_write_success(&us_buffer, got);
	return ParseBody(rq);
}
kev_result KAsyncFetchObject::PushBodyResult(KHttpRequest *rq, KGL_RESULT result)
{
	kev_result ret;
	switch (result) {
	case KGL_OK:
		ret = try_send_request(rq);
		if (KEV_HANDLED(ret)) {
			return ret;
		}
		return continueReadBody(rq);
	case STREAM_WRITE_END:
		assert(rq->ctx->connection_upgrade == false);
		expectDone(rq);
		return gate->f->push_body_finish(gate, rq, KGL_OK);
	default:
		if (rq->ctx->connection_upgrade) {
			return shutdown(rq);
		}
		return gate->f->push_body_finish(gate, rq, result);
	}
}
kev_result KAsyncFetchObject::ParseBody(KHttpRequest *rq)
{
	char *data = us_buffer.buf;
	int len = us_buffer.used;
	KGL_RESULT ret = ParseBody(rq, &data, &len);
	ks_save_point(&us_buffer,data, len);
	return PushBodyResult(rq,ret);	
}
kev_result KAsyncFetchObject::handleHttp2ReadHead(KHttpRequest *rq,int got)
{
	if (got<0) {
		return handleUpstreamError(rq,STATUS_SERVER_ERROR,"Cann't Send head to remote server",got);
	}
	assert(got==0);
	badStage = BadStage_SendSuccess;
	return readHeadSuccess(rq);
}
kev_result KAsyncFetchObject::sendHeadSuccess(KHttpRequest *rq)
{
	buffer->destroy();
	if (gate->f->get_post_left(gate,rq)!=0) {
		return readPost(rq);
	}
	return startReadHead(rq);
}
kev_result KAsyncFetchObject::startReadHead(KHttpRequest *rq)
{
	InitUpstreamBuffer();
	if (rq->ctx->connection_upgrade) {
		return readHeadSuccess(rq);
	}
	if (client->ReadHttpHeader(rq, resultUpstreamHttp2ReadHeader)) {
		return kev_ok;
	}
	client->SetNoDelay(false);
	return client->Read(rq, resultUpstreamReadHead, bufferUpstreamRead);
}
kev_result KAsyncFetchObject::handleSendHead(KHttpRequest *rq,int got)
{
	//printf("handleSendHead got=[%d]\n",got);
	if (got<=0) {
		return handleUpstreamError(rq,STATUS_SERVER_ERROR,"Cann't Send head to remote server",got);
	}
	badStage = BadStage_SendSuccess;
	if(buffer->readSuccess(got)){
		//continue send head
		return client->Write(rq, resultUpstreamSendHead, bufferUpstreamSendHead);
	}
	return sendHeadSuccess(rq);
}
kev_result KAsyncFetchObject::handleSendPost(KHttpRequest *rq,int got)
{
	if (got<=0) {
		return handleUpstreamError(rq,STATUS_SERVER_ERROR,"cann't send post data to remote server",got);
	}
	if (buffer->readSuccess(got)) {
		return client->Write(rq, resultUpstreamSendPost, bufferUpstreamSendPost);
	}
	//if (rq->ctx->connection_upgrade) {
		//client->Flush();
	//}
	buffer->destroy();
	//try to read post
	if (!rq->ctx->connection_upgrade && gate->f->get_post_left(gate,rq)==0) {
		return startReadHead(rq);
	}
	return readPost(rq);	
}
kev_result KAsyncFetchObject::shutdown(KHttpRequest *rq)
{
	if (!rq->sink->IsLocked() &&
		!client->IsLocked() &&
		!rq->ctx->pending) {
		return stageEndRequest(rq);
	}
	client->Shutdown();
	rq->sink->Shutdown();
	return kev_err;
}
kev_result KAsyncFetchObject::PostResult(KHttpRequest *rq,int got)
{
	//printf("handleReadPost got=[%d] protocol=[%d]\n",got,(int)rq->http_major);
	if (got == 0 && !rq->ctx->connection_upgrade) {
		if (chunk_post) {
			BuildChunkHeader();
			return sendPost(rq);
		}
		if (gate->f->get_post_left(gate, rq) == 0) {
			//Ҫ0ĻͱʾȡʧܣǶȡ
			client->WriteEnd();
			return startReadHead(rq);
		}
	}
	if (got <= 0) {
		if (rq->ctx->connection_upgrade) {
			return shutdown(rq);
		}
		return gate->f->push_message(gate, rq, KGL_MSG_ERROR, 500, "cann't read post");
		//return stageEndRequest(rq);
	}
	buffer->writeSuccess(got);
	if (chunk_post) {
		BuildChunkHeader();
	}
	return sendPost(rq);
}
void KAsyncFetchObject::BuildChunkHeader()
{
	int len = buffer->getLen();
	if (len == 0) {
		buffer->insert("0\r\n\r\n", 5);
		return;
	}
	char buf[32];
	int buf_len = snprintf(buf, sizeof(buf), "%x\r\n", len);
	buffer->insert(buf, buf_len);
	buffer->WSTRING("\r\n");
}
kev_result KAsyncFetchObject::readPost(KHttpRequest *rq)
{
	return gate->f->read_post(gate, rq);
}
kev_result KAsyncFetchObject::sendPost(KHttpRequest *rq)
{
	buildPost(rq);
	buffer->startRead();
	if (!rq->ctx->connection_upgrade) {
		read_hup(rq);
	}
	return client->Write(rq, resultUpstreamSendPost, bufferUpstreamSendPost);
}
kev_result KAsyncFetchObject::readHeadSuccess(KHttpRequest *rq)
{
	pop_header.first_body_time = kgl_current_sec;
	client->IsGood();
	if (rq->ctx->connection_upgrade) {
		rq->sink->SetNoDelay(true);
		client->SetNoDelay(true);
		int tmo = rq->sink->GetTimeOut();
		if (tmo < 5) {
			tmo = 5;
			rq->sink->SetTimeOut(tmo);
		}
		client->SetTimeOut(tmo);
		kev_result ret = gate->f->read_post(gate, rq);
		if (!KEV_AVAILABLE(ret)) {
			return ret;
		}
	}
	return PushHeaderFinished(rq);

}
kev_result KAsyncFetchObject::PushHeaderFinished(KHttpRequest *rq)
{
	if (rq->meth == METH_HEAD || pop_header.no_body) {
		//expected done
		expectDone(rq);
	}
	return gate->f->push_header_finish(gate, rq);
}
void KAsyncFetchObject::PushStatus(KHttpRequest *rq, int status_code)
{
	pop_header.status_send = 1;
	if (is_status_code_no_body(status_code)) {
		pop_header.no_body = 1;
	}
	gate->f->push_status(gate, rq, status_code);
}
KGL_RESULT KAsyncFetchObject::PushHeader(KHttpRequest *rq, const char *attr, int attr_len, const char *val, int val_len, bool request_line)
{
	if (request_line && pop_header.proto == Proto_http) {
		if (strncasecmp(attr, "HTTP/", 5) == 0) {
			const char *dot = strchr(attr + 5, '.');
			if (dot == NULL) {
				return KGL_EDATA_FORMAT;
			}
			uint8_t http_major = *(dot - 1) - 0x30;//major;
			uint8_t http_minor = *(dot + 1) - 0x30;//minor;
			if (http_major > 1) {
				rq->ctx->upstream_connection_keep_alive = true;
			} else if (http_major == 1 && http_minor == 1) {
				rq->ctx->upstream_connection_keep_alive = true;
			}
			PushStatus(rq, atoi(val));
			return KGL_OK;
		}
	}

	if (!strcasecmp(attr, "Keep-Alive")) {
		kassert(false);
		const char *data = strstr(val, "timeout=");
		if (data) {
			//ȷЧ2ʱ
			pop_header.keep_alive_time_out = atoi(data + 8) - 2;
		}
		return KGL_OK;
	}
	if (strcasecmp(attr, "Transfer-Encoding") == 0) {
		if (strcasecmp(val, "chunked") == 0) {
			gate = new_dechunk_push_gate(gate);
			return KGL_OK;
		}
	}
	if (!strcasecmp(attr, "Connection")) {
		KHttpFieldValue field(val);
		do {
			if (field.is2("keep-alive", 10)) {
				rq->ctx->upstream_connection_keep_alive = true;
			} else if (field.is2("close", 5)) {
				rq->ctx->upstream_connection_keep_alive = false;
			} else if (TEST(rq->flags, RQ_HAS_CONNECTION_UPGRADE) && field.is2("upgrade", 7)) {
				rq->ctx->connection_upgrade = true;
				rq->ctx->upstream_connection_keep_alive = false;
			}
		} while (field.next());
		return KGL_OK;
	}

	if (pop_header.proto == Proto_spdy && *attr == ':') {
		attr++;
		attr_len--;
		if (strcasecmp(attr, "version") == 0) {
			return KGL_OK;
		}
	}
	if (pop_header.proto != Proto_http) {
		if (!strcasecmp(attr, "Status")) {
			if (!pop_header.status_send) {
				PushStatus(rq, atoi(val));
			}
			return KGL_OK;
		}
		if (!strcasecmp(attr, "Location")) {
			if (!pop_header.status_send) {
				PushStatus(rq, STATUS_FOUND);
			}
		} else if (!strcasecmp(attr, "WWW-Authenticate")) {
			if (!pop_header.status_send) {
				PushStatus(rq, STATUS_UNAUTH);
			}
		}
	}
	return gate->f->push_unknow_header(gate, rq, attr, attr_len, val, val_len);
}
kev_result KAsyncFetchObject::handleReadHead(KHttpRequest *rq,int got)
{
	if (got<=0) {
		return handleUpstreamError(rq,STATUS_GATEWAY_TIMEOUT,"cann't recv head from remote server",got);
	}
	ks_write_success(&us_buffer, got);
	char *data = us_buffer.buf;
	int len = us_buffer.used;
	switch (ParseHeader(rq, &data, &len)) {
	case kgl_parse_finished:
		ks_save_point(&us_buffer, data, len);
		return readHeadSuccess(rq);
	case kgl_parse_continue:
		if (parser.header_len > MAX_HTTP_HEAD_SIZE) {
			return handleUpstreamError(rq, STATUS_GATEWAY_TIMEOUT, "upstream protocol header size is too big", got);
		}
		ks_save_point(&us_buffer, data, len);
		return client->Read(rq, resultUpstreamReadHead, bufferUpstreamRead);
	case kgl_parse_want_read:
		//ajpЭᷢһJK_AJP13_GET_BODY_CHUNKʱҪһյݰȥ
		return sendPost(rq);
	default:
		return handleUpstreamError(rq, STATUS_GATEWAY_TIMEOUT, "cann't parse upstream protocol", got);
	}
}
kgl_parse_result KAsyncFetchObject::ParseHeader(KHttpRequest *rq, char **data, int *len)
{
	khttp_parse_result rs;
	for (;;) {
		memset(&rs, 0, sizeof(rs));
		kgl_parse_result result = khttp_parse(&parser, data, len, &rs);
		switch (result) {
		case kgl_parse_error:
			return kgl_parse_error;
		case kgl_parse_continue:
			return kgl_parse_continue;
		case kgl_parse_success:
#if 0
			if (rs.is_first) {
				parser_ctx.StartParse(rq);
			}
#endif
			if (KGL_OK != PushHeader(rq, rs.attr, rs.attr_len, rs.val, rs.val_len, rs.request_line)) {
				return kgl_parse_error;
			}
			//printf("attr=[%s] val=[%s]\n", rs.attr, rs.val);
			break;
		case kgl_parse_finished:
			return kgl_parse_finished;
		}
	}
}
KGL_RESULT KAsyncFetchObject::ParseBody(KHttpRequest *rq, char **data, int *len)
{
	KGL_RESULT result = PushBody(rq, *data, *len);
	*len = 0;
	return result;
}
kev_result KAsyncFetchObject::handleUpstreamError(KHttpRequest *rq,int error,const char *msg,int last_got)
{
	kassert(client);
	int err = errno;
	pop_header.keep_alive_time_out = -1;
	if (!rq->ctx->read_huped) {
		if (client->IsNew()) {
			//newĲż.
			client->IsBad(badStage);
		}
		SET(rq->flags, RQ_UPSTREAM_ERROR);
	}
	if (!rq->ctx->connection_connect_proxy && rq->ctx->connection_upgrade) {
		return shutdown(rq);
	}
	if (rq->ctx->read_huped) {
		//client broken
		return gate->f->push_message(gate, rq, KGL_MSG_ERROR, error, msg);
	}	
	char *url = rq->url->getUrl();
	sockaddr_i *upstream_addr = client->GetAddr();
	char ips[MAXIPLEN];
	ksocket_sockaddr_ip(upstream_addr, ips, MAXIPLEN);
	klog(KLOG_WARNING, "rq=[%p] request=[%s %s] upstream=[%s:%d] self_port=[%d] error code=[%d],msg=[%s] errno=[%d %s],socket is %s,last_got=[%d].\n",
		rq,
		rq->getMethod(),
		url,
		ips,
		ksocket_addr_port(upstream_addr),
		client->GetSelfPort(),
		error,
		msg,
		err,
		strerror(err),
		(client->IsNew()?"new":"pool"),
		last_got
		);
	xfree(url);
	if (badStage != BadStage_SendSuccess && !client->IsNew()) {
		//use pool connectioin will try again
		return retryOpen(rq);
	}
	return gate->f->push_message(gate, rq, KGL_MSG_ERROR, error, msg);
}
