#include <map>
#include "KDsoAsyncFetchObject.h"
#include "KAccessDsoSupport.h"
#include "KDsoRedirect.h"
#include "KDsoNextUpstream.h"
#include "http.h"
#include "KCdnContainer.h"


kgl_async_context * kgl_get_async_context(KCONN cn)
{
	KDsoAsyncFetchObject *fo = (KDsoAsyncFetchObject *)cn;
	return &fo->ctx;
}
static kev_result check_next(KREQUEST r, KCONN cn, bool need_open)
{
	KDsoAsyncFetchObject *fo = (KDsoAsyncFetchObject *)cn;
	KHttpRequest *rq = (KHttpRequest *)r;
	if (fo->IsOpened()) {
		return fo->gate->f->push_message(fo->gate, rq, KGL_MSG_ERROR, 500, "fo is already opened");
	}
	kgl_list *pos = fo->queue.next;
	klist_remove(&fo->queue);
	fo->Close(rq);
	if (pos == &rq->fo) {
		assert(false);
		return send_error(rq, 500, "may have bug.");
	}
	KFetchObject *next = kgl_list_data(pos, KFetchObject, queue);
	return next->Check(rq);
}
static kev_result open_next(KREQUEST r, KCONN cn, kgl_push_gate *gate, const char *queue)
{
	KDsoAsyncFetchObject *fo = (KDsoAsyncFetchObject *)cn;
	KHttpRequest *rq = (KHttpRequest *)r;	
	if (!fo->IsOpened()) {
		return send_error(rq, 500, "fo is not opened");
	}
	return rq->OpenNextFetchObject(fo, gate, queue);	
}
static kev_result redirect(KREQUEST r, KCONN cn, kgl_push_gate *gate, const char *target, const char *queue_name)
{
	KDsoAsyncFetchObject *fo = (KDsoAsyncFetchObject *)cn;
	KHttpRequest *rq = (KHttpRequest *)r;
	if (!fo->IsOpened()) {
		return send_error(rq, 500, "fo is not opened");
	}
	kgl_list *pos = rq->fo.prev;
	if (pos == &rq->fo) {
		assert(false);
		if (gate) {
			gate->f->release(gate);
		}
		return stageEndRequest(rq);
	}
	KFetchObject *last_fo = kgl_list_data(pos, KFetchObject, queue);
	if (last_fo->IsOpened() || last_fo->filter || last_fo==static_cast<KFetchObject *>(fo)) {
		if (gate) {
			gate->f->release(gate);
		}
		return fo->gate->f->push_message(fo->gate,rq, KGL_MSG_ERROR, STATUS_SERVER_ERROR, last_fo->filter?"no final upstream":"final upstream already opened");
	}
	KFetchObject *redirect_fo = server_container->get(target);
	if (redirect_fo == NULL) {
		if (gate) {
			gate->f->release(gate);
		}
		return fo->gate->f->push_message(fo->gate, rq, KGL_MSG_ERROR, STATUS_SERVER_ERROR, "cann't find target");
	}
	klist_remove(pos);
	delete last_fo;
	rq->AppendFetchObject(redirect_fo);
	return rq->OpenNextFetchObject(fo, gate, queue_name);
}
static bool is_next_opened(KREQUEST r, KCONN cn)
{
	KDsoAsyncFetchObject *fo = (KDsoAsyncFetchObject *)cn;
	KFetchObject *next = ((KHttpRequest *)r)->GetNextFetchObject(fo);
	if (next == NULL) {
		return false;
	}
	return next->IsOpened();
}
static kev_result try_write(KREQUEST r, KCONN  cn)
{
	KHttpRequest *rq = (KHttpRequest *)r;
	KDsoAsyncFetchObject *fo = (KDsoAsyncFetchObject *)cn;
	if (!fo->IsOpened()) {
		return send_error(rq, 500, "fo is not opened");
	}
	return try_send_request(rq);
}
static KGL_RESULT support_function (
	KREQUEST r,
	KCONN      cn,
	KF_REQ_TYPE  req,
	PVOID    data,
	PVOID *ret)
{
	KHttpRequest *rq = (KHttpRequest *)r;
	return base_support_function(rq, req, data, ret);
}

static kgl_async_context_function  async_context_function = {
	(kgl_get_variable_f)get_request_variable,
	check_next,
	open_next,
	redirect,
	is_next_opened,
	try_write,
	support_function
};

KDsoAsyncFetchObject::~KDsoAsyncFetchObject()
{
	if (ctx.module) {
		kassert(brd!=NULL);
		GetAsyncUpstream()->free_ctx(ctx.module);
	}
}
void KDsoAsyncFetchObject::Close(KHttpRequest *rq)
{
	if (ctx.f) {
		kgl_async_upstream *async_us = GetAsyncUpstream();
		kassert(this == ctx.cn);
		async_us->close(rq, &ctx);
		KFetchObject::Close(rq);
		ctx.f = NULL;
	}
}
void KDsoAsyncFetchObject::init()
{
	ctx.cn = this;	
	ctx.us = &next_upstream;
	ctx.f = &async_context_function;
	if (ctx.module == NULL) {
		ctx.module = GetAsyncUpstream()->create_ctx();
	}
}
kev_result KDsoAsyncFetchObject::Open(KHttpRequest *rq, kgl_push_gate *gate)
{
	KFetchObject::Open(rq, gate);
	kassert(brd);
	if (ctx.f == NULL) {
		init();
	}
	ctx.gate = this->gate;
	return GetAsyncUpstream()->open(rq, &ctx);
}
kev_result KDsoAsyncFetchObject::Read(KHttpRequest *rq)
{
	return GetAsyncUpstream()->read(rq, &ctx);
}

kev_result KDsoAsyncFetchObject::PostResult(KHttpRequest *rq, int got)
{
	kgl_async_upstream *async_us = GetAsyncUpstream();
	if (async_us->post_result) {
		return async_us->post_result(rq, &ctx, got);
	}
	return KFetchObject::PostResult(rq, got);
}
char *KDsoAsyncFetchObject::GetPostBuffer(KHttpRequest *rq, int *len)
{
	kgl_async_upstream *async_us = GetAsyncUpstream();
	if (async_us->get_post_buffer) {
		return async_us->get_post_buffer(rq, &ctx, len);
	}
	return KFetchObject::GetPostBuffer(rq, len);
}
kev_result KDsoAsyncFetchObject::Check(KHttpRequest *rq)
{
	if (ctx.f == NULL) {
		init();
	}
	ctx.gate = get_check_push_gate();
	kgl_async_upstream *async_us = GetAsyncUpstream();
	if (async_us->check) {
		return async_us->check(rq, &ctx);
	}
	return KFetchObject::Check(rq);
}
