/*
 * KFetchObject.cpp
 *
 *  Created on: 2010-4-19
 *      Author: keengo
 * Copyright (c) 2010, NanChang BangTeng Inc
 * All Rights Reserved.
 *
 * You may use the Software for free for non-commercial use
 * under the License Restrictions.
 *
 * You may modify the source code(if being provieded) or interface
 * of the Software under the License Restrictions.
 *
 * You may use the Software for commercial use after purchasing the
 * commercial license.Moreover, according to the license you purchased
 * you may get specified term, manner and content of technical
 * support from NanChang BangTeng Inc
 *
 * See COPYING file for detail.
 */
#include "KFetchObject.h"
#include "kmalloc.h"
#include "KAsyncFetchObject.h"
#include "http.h"
#include "KHttpTransfer.h"
KFetchObject::~KFetchObject()
{
	if (brd) {
		brd->release();
	}
	kassert(gate == NULL);
	if (gate) {
		gate->f->release(gate);
	}
}
/*
Ѿupstreamݣtrue,ȡfalse򲻼ʾѾݿԷrq
*/
kev_result KFetchObject::pushHttpBody(KHttpRequest *rq,const char *buf,int len)
{
	KGL_RESULT result = PushBody(rq, buf, len);
	switch (result) {
	case KGL_OK:
		return try_send_request(rq);
	case STREAM_WRITE_END:
		kassert(false);//never to here
	default:
		kassert(!rq->ctx->connection_upgrade);
		return gate->f->push_body_finish(gate, rq, result);
	}	
}
kev_result KFetchObject::Read(KHttpRequest *rq)
{
	if (queue.next != &rq->fo) {
		KFetchObject *next = kgl_list_data(queue.next, KFetchObject, queue);
		return next->Read(rq);
	}
	return kev_err;
}
kev_result KFetchObject::Check(KHttpRequest *rq)
{
	KFetchObject *next = rq->GetNextFetchObject(this);
	if (next) {
		return next->Check(rq);
	}
	//ܵ
	assert(false);
	return kev_err;
}
kev_result KFetchObject::Open(KHttpRequest *rq,kgl_push_gate *gate)
{
	this->opened = 1;
	if (this->gate && gate) {
		assert(false);
		this->gate->f->release(this->gate);
	}
	if (gate) {
		this->gate = gate;
	}
	if (this->gate == NULL && !isSync()) {
		this->gate = new_default_push_gate();
	}
	return kev_err;
}
bool KFetchObject::NeedTempFile(bool upload, KHttpRequest *rq)
{
	if (!upload) {
		return true;
	}
	return rq->content_length == -1;
}
KGL_RESULT KFetchObject::PushBody(KHttpRequest *rq, const char *buf, int len)
{
	if (!rq->ctx->connection_upgrade && rq->ctx->know_length) {
		len = (int)MIN(rq->ctx->left_read, (INT64)len);
		rq->ctx->left_read -= len;
	}
	return gate->f->push_body(gate, rq, buf, len);
}
void KFetchObject::Close(KHttpRequest *rq)
{
	this->opened = 0;
	if (gate) {
		gate->f->release(gate);
		gate = NULL;
	}
}
bool KFetchObject::IsLocked(KHttpRequest *rq)
{
	KUpstream *us = GetUpstream();
	if (us) {
		if (us->IsLocked()) {
			return true;
		}
	}
	KFetchObject *next = rq->GetNextFetchObject(this);
	if (next) {
		return next->IsLocked(rq);
	}
	return false;
}
kev_result KFetchObject::PostResult(KHttpRequest *rq, int got)
{
	KFetchObject *next = rq->GetNextFetchObject(this);
	if (next) {
		return next->PostResult(rq, got);
	}
	assert(false);
	return stageEndRequest(rq);
}
char *KFetchObject::GetPostBuffer(KHttpRequest *rq, int *size)
{
	KFetchObject *next = rq->GetNextFetchObject(this);
	if (next) {
		return next->GetPostBuffer(rq, size);
	}
	assert(false);
	return 0;
}
void KFetchObject::Shutdown(KHttpRequest *rq)
{
	KUpstream *us = GetUpstream();
	if (us) {
		us->Shutdown();
	}
	KFetchObject *next = rq->GetNextFetchObject(this);
	if (next) {
		next->Shutdown(rq);
	}
}
kev_result KRefFetchObject::Check(KHttpRequest *rq)
{
	assert(rq->GetNextFetchObject(this) == NULL);
	klist_remove(&queue);
	if (this->fo) {
		rq->AppendFetchObject(this->fo);
		this->fo = NULL;
	}
	kev_result ret = async_http_start(rq);
	delete this;
	return ret;
}
kev_result KRefFetchObject::Open(KHttpRequest *rq, kgl_push_gate *gate)
{
	//ܵõ
	assert(false);
	KFetchObject::Open(rq, gate);
	//ԴһԴ
	assert(rq->GetNextFetchObject(this) == NULL);
	return kev_err;
}
