#include "KPushGate.h"
#include "http.h"
#include "ksapi.h"
#include "KDechunkEngine.h"
#include "KBufferFetchObject.h"
#include "KVectorBufferFetchObject.h"

kev_result result_read_post(KOPAQUE data, void *arg, int got)
{
	KHttpRequest *rq = (KHttpRequest *)data;
	KFetchObject *fo = (KFetchObject *)arg;
	if (got == 0 && rq->left_read == -1 && !rq->ctx->connection_upgrade) {
		rq->left_read = 0;
		return fo->PostResult(rq, got);
	}
	if (got <= 0) {
		return fo->PostResult(rq, got);
	}
	rq->AddUpFlow(got);
	if (!rq->ctx->connection_upgrade && rq->left_read != -1) {
		rq->left_read -= got;
	}
	return fo->PostResult(rq, got);
}
int buffer_read_post(KOPAQUE data, void *arg, LPWSABUF buf, int bc)
{
	KHttpRequest *rq = (KHttpRequest *)data;
	KFetchObject *fo = (KFetchObject *)arg;
	int len;
	char *str = fo->GetPostBuffer(rq, &len);
	if (!rq->ctx->connection_upgrade && rq->left_read >= 0) {
		len = (int)(MIN((INT64)len, rq->left_read));
		assert(len > 0);
	}
	buf[0].iov_base = str;
	buf[0].iov_len = len;
	return 1;
}
struct kgl_default_push_gate : public kgl_push_gate {
	KHttpResponseParser parser_ctx;
};
struct kgl_forward_push_gate : public kgl_push_gate {
	kgl_push_gate *down_gate;
	bool release_down_gate;
};
struct kgl_dechunk_push_gate : public kgl_forward_push_gate {
	KDechunkEngine engine;
};
static int64_t default_get_post_left(kgl_push_gate *gate, KREQUEST r)
{
	KHttpRequest *rq = (KHttpRequest *)r;
	return rq->left_read;
}
static kev_result default_read_post(kgl_push_gate *gate, KREQUEST r)
{
	KHttpRequest *rq = (KHttpRequest *)r;
	KFetchObject *fo = rq->GetFetchObject();
	assert(fo != NULL);
	return rq->Read(fo, result_read_post, buffer_read_post);
}
static void push_status(kgl_push_gate *gate, KREQUEST rq, int status_code)
{
	((KHttpRequest *)rq)->ctx->obj->data->status_code = status_code;
}
static KGL_RESULT forward_push_unknow_header(kgl_push_gate *gate, KREQUEST r, kgl_header_type attr, const char *val, hlen_t val_len)
{
	return gate->f->push_unknow_header(gate, r, kgl_header_type_string[attr].data, (hlen_t)kgl_header_type_string[attr].len, val, val_len);
}
static KGL_RESULT push_header(kgl_push_gate *gate, KREQUEST r, kgl_header_type attr, const char *val, hlen_t val_len)
{
	kgl_default_push_gate *g = (kgl_default_push_gate *)gate;
	KHttpRequest *rq = (KHttpRequest *)r;
	KHttpObject *obj = rq->ctx->obj;
	if (val_len == 0) {
		switch (attr) {
		case  kgl_header_content_length:
		{
			INT64 *content_length = (INT64*)val;
			kassert(*content_length >= 0);
			obj->index.content_length = *content_length;
			SET(obj->index.flags, ANSW_HAS_CONTENT_LENGTH);
			return KGL_OK;
		}
		case kgl_header_last_modified:
		{
			obj->index.last_modified = *(time_t *)val;
			char tmp_buf[42];
			mk1123time(obj->index.last_modified, tmp_buf, 41);
			SET(obj->index.flags, ANSW_LAST_MODIFIED);
			g->parser_ctx.AddHeader(kgl_header_type_string[attr].data, (hlen_t)kgl_header_type_string[attr].len, tmp_buf, 29);
			return KGL_OK;
		}
		default:
		{
			return KGL_ENOT_SUPPORT;
		}
		}
	}
	return forward_push_unknow_header(gate, r, attr, val, val_len);
}
static KGL_RESULT push_unknow_header(kgl_push_gate *gate, KREQUEST rq, const char *attr, hlen_t attr_len, const char *val, hlen_t val_len)
{
	kgl_default_push_gate *g = (kgl_default_push_gate *)gate;
	if (g->parser_ctx.ParseHeader((KHttpRequest *)rq, attr, attr_len, val, val_len, false)) {
		return KGL_OK;
	}
	return KGL_EDATA_FORMAT;
}
static kev_result push_header_finish(kgl_push_gate *gate, KREQUEST r)
{
	kgl_default_push_gate *g = (kgl_default_push_gate *)gate;
	KHttpRequest *rq = (KHttpRequest *)r;
	g->parser_ctx.EndParse(rq);
	return handleUpstreamRecvedHead(rq);
}
static KGL_RESULT push_body(kgl_push_gate *gate, KREQUEST r, const char *buf, int len)
{
	KHttpRequest *rq = (KHttpRequest *)r;
	kassert(rq->ctx->st);
	return rq->ctx->st->write_all(rq, buf, len);
}
static kev_result push_body_finish(kgl_push_gate *gate, KREQUEST rq, KGL_RESULT result)
{
	return stage_rdata_end((KHttpRequest *)rq, result);
}
static kev_result default_push_message(kgl_push_gate *gate, KREQUEST r, KGL_MSG_TYPE msg_type, int32_t msg_flag, const char *msg)
{
	KHttpRequest *rq = (KHttpRequest *)r;
	if (msg_type == KGL_MSG_ERROR) {
		return handleError(rq, msg_flag, msg);
	}
	if (TEST(rq->flags, RQ_HAS_SEND_HEADER)) {			
		return stageEndRequest(rq);
	}
	rq->closeFetchObject();
	KFetchObject *fo = NULL;
	int64_t len;
	switch (msg_type) {
	case KGL_MSG_RAW:
	{
		len = (int64_t)msg_flag;
		if (len > 0) {
			kbuf *buf = (kbuf *)rq->alloc_request_memory(sizeof(kbuf));
			memset(buf, 0, sizeof(kbuf));
			buf->data = (char *)msg;
			buf->used = (int)len;
			fo = new KBufferFetchObject(buf, 0, msg_flag, rq->pool);
		}
		break;
	}
	case KGL_MSG_VECTOR:
	{
		KVectorBufferFetchObject *vfo = new KVectorBufferFetchObject((WSABUF *)msg, (int)msg_flag);
		len = vfo->GetLen();
		fo = vfo;
		break;
	}
	default:
		break;
	}	
	if (rq->status_code == 0) {
		rq->responseStatus(200);
	}
	char len_str[INT2STRING_LEN];
	int len2 = snprintf(len_str, INT2STRING_LEN - 1, INT64_FORMAT, len);
	rq->responseHeader(kgl_expand_string("Content-Length"), len_str, len2);
	rq->responseConnection();
	rq->startResponseBody(len);
	if (fo) {
		rq->AppendFetchObject(fo);
		return fo->Open(rq, NULL);
	}
	return stageEndRequest(rq);
	
}
static void release(kgl_push_gate *gate)
{
	kgl_default_push_gate *g = (kgl_default_push_gate *)gate;
	delete g;
}

static kgl_push_gate_function default_push_gate_function = {
	default_get_post_left,
	default_read_post,
	push_status,
	push_header,
	push_unknow_header,
	push_header_finish,
	push_body,
	push_body_finish,
	default_push_message,
	release
};
kgl_push_gate *new_default_push_gate()
{
	kgl_default_push_gate *gate = new kgl_default_push_gate;
	gate->f = &default_push_gate_function;
	return gate;
}
static KGL_RESULT dechunk_push_body(kgl_push_gate *gate, KREQUEST r, const char *buf, int len)
{
	kgl_dechunk_push_gate *g = (kgl_dechunk_push_gate *)gate;
	const char *piece = NULL;
	int piece_length;
	while (len > 0) {
		dechunk_status status = g->engine.dechunk(&buf, len, &piece, piece_length);
		switch (status) {
		case dechunk_failed:
			return STREAM_WRITE_FAILED;
		case dechunk_end:
			return STREAM_WRITE_END;
		case dechunk_continue:
			if (piece != NULL) {
				KGL_RESULT result = g->down_gate->f->push_body(g->down_gate, r, piece, piece_length);
				if (result != KGL_OK) {
					return result;
				}
			}
			return STREAM_WRITE_SUCCESS;
		default:
			assert(piece != NULL);
			KGL_RESULT result = g->down_gate->f->push_body(g->down_gate, r, piece, piece_length);
			if (result != KGL_OK) {
				return result;
			}
		}
	}
	return STREAM_WRITE_SUCCESS;
}
int64_t forward_get_post_left(kgl_push_gate *gate, KREQUEST r)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->get_post_left(g->down_gate, r);
}
kev_result forward_read_post(kgl_push_gate *gate, KREQUEST r)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->read_post(g->down_gate, r);
}
KGL_RESULT forward_push_header(kgl_push_gate *gate, KREQUEST r, kgl_header_type attr, const char *val, hlen_t val_len)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_header(g->down_gate, r, attr,val,val_len);
}
KGL_RESULT forward_push_unknow_header(kgl_push_gate *gate, KREQUEST r, const char *attr, hlen_t attr_len, const char *val, hlen_t val_len)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_unknow_header(g->down_gate, r, attr, attr_len, val, val_len);
}
void forward_push_status(kgl_push_gate *gate, KREQUEST r, int status_code)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_status(g->down_gate, r, status_code);
}
static KGL_RESULT forward_push_body(kgl_push_gate *gate, KREQUEST r, const char *buf, int len)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_body(g->down_gate, r, buf, len);
}
static kev_result forward_push_header_finish(kgl_push_gate *gate, KREQUEST r)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_header_finish(g->down_gate, r);
}
static kev_result forward_push_body_finish(kgl_push_gate *gate, KREQUEST rq, KGL_RESULT result)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_body_finish(g->down_gate, rq, result);
}
kev_result forward_handle_error(kgl_push_gate *gate, KREQUEST rq, KGL_MSG_TYPE type, int32_t msg_flag, const char *msg)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	return g->down_gate->f->push_message(g->down_gate, rq, type, msg_flag, msg);
}
void forward_push_gate_release(kgl_push_gate *gate)
{
	kgl_forward_push_gate *g = (kgl_forward_push_gate *)gate;
	if (g->release_down_gate) {
		g->down_gate->f->release(g->down_gate);
	}
	delete g;
}
static void dechunk_release(kgl_push_gate *gate)
{
	kgl_dechunk_push_gate *g = (kgl_dechunk_push_gate *)gate;
	g->down_gate->f->release(g->down_gate);
	delete g;
}
static kgl_push_gate_function dechunk_push_gate_function = {
	forward_get_post_left,
	forward_read_post,
	forward_push_status,
	forward_push_header,
	forward_push_unknow_header,
	forward_push_header_finish,
	dechunk_push_body,
	forward_push_body_finish,
	forward_handle_error,
	dechunk_release
};
kgl_push_gate *new_dechunk_push_gate(kgl_push_gate *down_gate)
{
	kgl_dechunk_push_gate *gate = new kgl_dechunk_push_gate;
	gate->f = &dechunk_push_gate_function;
	gate->down_gate = down_gate;
	gate->release_down_gate = true;
	return gate;
}
kgl_push_gate_function forward_push_gate_function = {
	forward_get_post_left,
	forward_read_post,
	forward_push_status,
	forward_push_header,
	forward_push_unknow_header,
	forward_push_header_finish,
	forward_push_body,
	forward_push_body_finish,
	forward_handle_error,
	forward_push_gate_release
};
kgl_push_gate *new_forward_push_gate(kgl_push_gate *down_gate, bool release_down_gate)
{
	kgl_forward_push_gate *gate = new kgl_forward_push_gate;
	gate->f = &forward_push_gate_function;
	gate->down_gate = down_gate;
	gate->release_down_gate = release_down_gate;
	return gate;
}
static void check_push_status(kgl_push_gate *gate, KREQUEST rq, int status_code)
{
	((KHttpRequest *)rq)->responseStatus(status_code);
}
static KGL_RESULT check_push_unknow_header(kgl_push_gate *gate, KREQUEST rq, const char *attr, hlen_t attr_len, const char *val, hlen_t val_len)
{
	if (((KHttpRequest *)rq)->responseHeader(attr, attr_len, val, val_len)) {
		return KGL_OK;
	}
	return KGL_EINVALID_PARAMETER;
}
static kev_result check_push_header_finish(kgl_push_gate *gate, KREQUEST r)
{
	return send_error((KHttpRequest *)r, 500, "do not support in check process");
}
static KGL_RESULT check_push_body(kgl_push_gate *gate, KREQUEST r, const char *buf, int len)
{
	return KGL_ENOT_SUPPORT;
}
static kev_result check_push_body_finish(kgl_push_gate *gate, KREQUEST r, KGL_RESULT result)
{
	return send_error((KHttpRequest *)r, 500, "do not support in check process");
}
static void check_release(kgl_push_gate *gate)
{
}
static kgl_push_gate_function check_push_gate_function = {
	default_get_post_left,
	default_read_post,
	check_push_status,
	forward_push_unknow_header,
	check_push_unknow_header,
	check_push_header_finish,
	check_push_body,
	check_push_body_finish,
	default_push_message,
	check_release,
};
static kgl_push_gate check_push_gate = {

	&check_push_gate_function,
};
kgl_push_gate *get_check_push_gate()
{
	return &check_push_gate;
}
