#include <string.h>
#include <string>
#include "KRequestQueue.h"
#include "kthread.h"
#include "kselectable.h"
#include "http.h"
#include "kselector.h"
#include "kmalloc.h"
#ifdef ENABLE_REQUEST_QUEUE
KRequestQueue globalRequestQueue;
inline kev_result stage_async_need_queue(KQueueParam *qp)
{
	KHttpRequest *rq = qp->rq;
	rq->begin_time_msec = kgl_current_msec;
	rq->sink->RemoveSync();
	kev_result ret = qp->fo->Open(rq,NULL);
	delete qp;
	return ret;
}
kev_result resultQueuedRequestTimeOut(void *data, void *arg,int got)
{	
	KHttpRequest *rq = (KHttpRequest *)arg;
	if (got<0) {
		SET(rq->flags,RQ_CONNECTION_CLOSE);
		return stageEndRequest(rq);
	}
	rq->closeFetchObject();
	return send_error(rq, STATUS_SERVICE_UNAVAILABLE, "Server is busy");
}
bool checkQueuedRequestTimeOut(KQueueParam *qp)
{
	if (kgl_current_msec - qp->rq->begin_time_msec > conf.time_out * 1000) {
		CLR(qp->rq->flags,RQ_SYNC);
		kgl_selector_module.next(qp->rq->sink->GetSelector(),NULL, resultQueuedRequestTimeOut, qp->rq, 0);
		delete qp;
		return true;
	}
	return false;
}
kev_result resultAsyncNextRequest(KOPAQUE data, void *arg,int got)
{
	KQueueParam *rq = (KQueueParam *)arg;
	return stage_async_need_queue(rq);
}

void async_queue_destroy(KRequestQueue *queue)
{
	assert(queue);
	KQueueParam *qp = queue->getQueue();
	if (qp == NULL || checkQueuedRequestTimeOut(qp)) {
		return;
	}
	assert(qp->rq->queue == queue);
	if (unlikely(qp->rq->IsSync())) {
		if(!kthread_pool_start(stage_sync, qp->rq)){
			qp->rq->queue = NULL;
			SET(qp->rq->flags,RQ_CONNECTION_CLOSE);
			queue->decWorker();
			stageEndRequest(qp->rq);			
		}
		delete qp;
		return;
	}
	kgl_selector_module.next(qp->rq->sink->GetSelector(),NULL, resultAsyncNextRequest, qp, 0);	
}
KTHREAD_FUNCTION thread_queue(void *param)
{
	//debug("thread_queue process %p...\n",param);
	KQueueParam *qp = (KQueueParam *)param;
	for(;;){
		KRequestQueue *queue = qp->rq->queue;
		assert(queue);
		if (unlikely(qp->rq->IsSync())) {
			qp->rq->queue = NULL;
			stage_sync(qp->rq);
			delete qp;
			qp = queue->getQueue();
			queue->release();
			if (qp == NULL || checkQueuedRequestTimeOut(qp)) {
				break;
			}
		} else {
			stage_async_need_queue(qp);
			break;
		}	
	}

	KTHREAD_RETURN;
}
KRequestQueue::KRequestQueue()
{
	cur_worker = 0;
	max_worker = 0;
	max_queue = 0;
}
KRequestQueue::~KRequestQueue()
{
	assert(queue.empty());
}
void KRequestQueue::set(unsigned max_worker,unsigned max_queue)
{
	this->max_worker = max_worker;
	this->max_queue = max_queue;
}
bool KRequestQueue::startDirect(KHttpRequest *rq, KFetchObject *fo)
{
	KQueueParam *qp = new KQueueParam;
	qp->rq = rq;
	qp->fo = fo;
	if(likely(!TEST(rq->flags,RQ_SYNC))){
		stage_async_need_queue(qp);
		return true;
	}
	if (!kthread_pool_start(thread_queue, qp)) {
		delete qp;
		return false;
	}
	return true;
}
bool KRequestQueue::start(KHttpRequest *rq, KFetchObject *fo)
{
	bool result = false;
	SET(rq->flags,RQ_QUEUED);
	bool startResult = false;
	refsLock.Lock();
	if (rq->queue != NULL) {
		assert(rq->queue == this);
		rq->queue = NULL;
		refs --;
	}
	if (max_worker == 0 || cur_worker < max_worker) {
		rq->queue = this;
		refs ++;
		cur_worker ++;
		startResult = true;
		rq->ctx->queue_handled = true;
	} else {
		if(max_queue == 0 || queue.size() < max_queue){
			//printf("add to queue..............\n");
#ifndef NDEBUG
			klog(KLOG_DEBUG,"add %p to queue\n",rq);
#endif
			rq->setState(STATE_QUEUE);
			KQueueParam *qp = new KQueueParam;
			qp->rq = rq;
			qp->fo = fo;
			queue.push_back(qp);
			result = true;
			rq->ctx->queue_handled = true;
			if(!TEST(rq->flags,RQ_SYNC)){
				//in queue
				rq->sink->AddSync();
			}
		}
	}
	refsLock.Unlock();
	if (startResult) {
		result = startDirect(rq,fo);
		if (!result) {
			rq->queue = NULL;
			rq->ctx->queue_handled = false;
			refsLock.Lock();
			refs --;
			cur_worker --;
			refsLock.Unlock();
		}
	}
	return result;
}
KQueueParam *KRequestQueue::getQueue()
{
	KQueueParam *context;
	refsLock.Lock();
	if ((max_worker == 0 || cur_worker <= max_worker) && queue.size()>0) {
		context = *(queue.begin());
		queue.pop_front();
		assert(context);
		context->rq->queue = this;
		refs++;
		//debug("***********************get from queue size=%d\n",queue.size());
	} else {
		context = NULL;
		cur_worker --;
	}
	refsLock.Unlock();
	return context;
}
#endif
