#include <stdio.h>
#include "KTsUpstream.h"
#include "kselectable.h"
#include "kconnection.h"
extern void http2_header_callback(KOPAQUE data, void *arg, const char *attr, int attr_len, const char *val, int val_len);
static void ts_header_callback(KOPAQUE data, void *arg, const char *attr, int attr_len, const char *val, int val_len)
{
	KTsUpstream *ts = (KTsUpstream *)arg;
	kgl_pool_t *pool = ts->us->GetPool();
	KHttpHeader *header = new_pool_http_header(pool, attr, attr_len, val, val_len);
	if (ts->header) {
		ts->last_header->next = header;
		ts->last_header = header;
	} else {
		ts->header = header;
		ts->last_header = header;
	}
	return;
}
static kev_result next_ts_destroy(KOPAQUE data, void *arg, int got)
{
	KUpstream *us = (KUpstream *)arg;
	us->Destroy();
	return kev_ok;
}
static kev_result next_ts_gc(KOPAQUE data, void *arg, int got)
{
	KUpstream *us = (KUpstream *)arg;
	us->Gc(got,us->expire_time);
	return kev_ok;
}
static kev_result next_ts_read_header(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	return ts_us->NextReadHeader(data);
}
static kev_result next_ts_read(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	return ts_us->NextRead();
}
static kev_result next_ts_write(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	return ts_us->NextWrite();
}
static kev_result result_ts_read2(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	kassert(kselector_is_same_thread(ts_us->main_selector));
	kassert(ts_us->read_lock);
	ts_us->read_lock = 0;
	while (ts_us->header) {
		kassert(ts_us->stack[OP_READ].header == http2_header_callback);
		ts_us->stack[OP_READ].header(data, ts_us->stack[OP_READ].arg, ts_us->header->attr, ts_us->header->attr_len, ts_us->header->val, ts_us->header->val_len);
		ts_us->header = ts_us->header->next;
	}
	return ts_us->stack[OP_READ].result(data, ts_us->stack[OP_READ].arg, got);
}
static kev_result result_ts_read(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	kgl_selector_module.next(ts_us->main_selector, data, result_ts_read2, arg, got);
	return kev_ok;
}
static kev_result result_ts_write2(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	kassert(kselector_is_same_thread(ts_us->main_selector));
	kassert(ts_us->write_lock);
	ts_us->write_lock = 0;
	return ts_us->stack[OP_WRITE].result(data, ts_us->stack[OP_WRITE].arg, got);
}
static kev_result result_ts_write(KOPAQUE data, void *arg, int got)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	kgl_selector_module.next(ts_us->main_selector, data, result_ts_write2, arg, got);
	return kev_ok;
}
static int buffer_ts_read(KOPAQUE data, void *arg, LPWSABUF buf, int bc)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	return ts_us->stack[OP_READ].buffer(buf, bc);
}
static int buffer_ts_write(KOPAQUE data, void *arg, LPWSABUF buf, int bc)
{
	KTsUpstream *ts_us = (KTsUpstream *)arg;
	return ts_us->stack[OP_WRITE].buffer(buf, bc);
}
bool KTsUpstream::SetHttp2Parser(void *arg, kgl_header_callback header)
{
	assert(us->IsMultiStream());
	memset(&stack[OP_READ], 0, sizeof(KTsUpstreamStack));
	stack[OP_READ].arg = arg;
	stack[OP_READ].header = header;
	http_header_parser = 1;
	return true;
}
void KTsUpstream::TrySetHttp2Parser()
{
	if (!http_header_parser) {
		return;
	}
	http_header_parser = 0;
	us->SetHttp2Parser(this, ts_header_callback);
}
kev_result KTsUpstream::NextReadHeader(KOPAQUE data)
{
	TrySetHttp2Parser();
	if (write_end) {
		write_end = 0;
		us->WriteEnd();
	}
	if (!us->ReadHttpHeader(this,result_ts_read)) {
		result_ts_read(data,this, -1);
	}
	return kev_ok;
}
bool KTsUpstream::ReadHttpHeader(void *arg,  result_callback result)
{
	if (!us->IsMultiStream()) {
		return false;
	}
	assert(stack[OP_READ].header == http2_header_callback);
	assert(stack[OP_READ].arg == arg);
	kassert(read_lock == 0);
	read_lock = 1;
	stack[OP_READ].arg = arg;
	stack[OP_READ].result = result;
	selectable_next(&us->GetConnection()->st, next_ts_read_header, this, 0);
	return true;
}
kev_result KTsUpstream::Connect(void *arg, result_callback result)
{
	fprintf(stderr, "never goto here.\n");
	kassert(false);
	return kev_err;
}
kev_result KTsUpstream::NextRead()
{
	if (shutdown) {
		us->Shutdown();
	}
	return us->Read(this, result_ts_read, buffer_ts_read);
}
kev_result KTsUpstream::NextWrite()
{
	TrySetHttp2Parser();
	if (shutdown) {
		us->Shutdown();
	}
	return us->Write(this, result_ts_write, buffer_ts_write);
}
kev_result KTsUpstream::Read(void *arg, result_callback result, buffer_callback buffer)
{
	//printf("ts_upstream read.\n");
	kassert(kselector_is_same_thread(main_selector));
	kassert(read_lock == 0);
	kselectable *st = &us->GetConnection()->st;
	read_lock = 1;
	stack[OP_READ].arg = arg;
	stack[OP_READ].result = result;
	stack[OP_READ].bc = buffer(us->GetOpaque(), arg, stack[OP_READ].buf, TS_UPSTREAM_BUFFER_COUNT);
	selectable_next(st, next_ts_read, this, 0);
	return kev_ok;
}
kev_result KTsUpstream::Write(void *arg, result_callback result, buffer_callback buffer)
{
	//printf("ts_upstream write.\n");
	kassert(kselector_is_same_thread(main_selector));
	kassert(write_lock == 0);
	kselectable *st = &us->GetConnection()->st;
	write_lock = 1;
	stack[OP_WRITE].arg = arg;
	stack[OP_WRITE].result = result;
	stack[OP_WRITE].bc = buffer(us->GetOpaque(), arg, stack[OP_WRITE].buf, TS_UPSTREAM_BUFFER_COUNT);
	selectable_next(st, next_ts_write, this, 0);
	return kev_ok;
}
void KTsUpstream::WriteEnd()
{
	if (us->IsMultiStream()) {
		write_end = 1;
		return;
	}
	us->WriteEnd();
}
void KTsUpstream::Shutdown()
{
	if (us->IsMultiStream()) {
		//http2shutdown
		//printf("*****************use soft shutdown.\n");
		shutdown = 1;
		return;
	}
	us->Shutdown();
}
void KTsUpstream::Destroy()
{
	fprintf(stderr, "never goto here.\n");
	kassert(false);
	if (us) {
		selectable_next(&us->GetConnection()->st, next_ts_destroy, us, 0);
		us = NULL;
	}
	delete this;
}
bool KTsUpstream::IsLocked()
{
	return read_lock || write_lock;
}
void KTsUpstream::Gc(int life_time,time_t base_time)
{
	us->expire_time = base_time;
	selectable_next(&us->GetConnection()->st, next_ts_gc, us, life_time);
	us = NULL;
	delete this;
}
