/*
 * KVirtualProcess.h
 *
 *  Created on: 2010-10-24
 *      Author: keengo
 */

#ifndef KVIRTUALPROCESS_H
#define KVIRTUALPROCESS_H
#include <sstream>
#include "global.h"
#include "KExtendProgram.h"
#include "KTcpUpstream.h"
#include "KPoolableSocketContainer.h"
#include "KHttpRequest.h"
#include "KProcess.h"
#include "time_utils.h"

enum VProcess_Status
{	
	VProcess_Poweroff,	
	VProcess_Inprogress,
	VProcess_Poweron,
	VProcess_Close
};
struct KVirtualHostProcessQueue
{
	KHttpRequest *rq;
	KAsyncFetchObject *fo;
};
/*
 * һģ
 */
class KVirtualHostProcess : public KPoolableSocketContainer{
public:
	KVirtualHostProcess() {
		//closed = false;
		lastActive = kgl_current_sec;
		idleTime = 0;
		max_request = 0;
		max_error_count = 0;
		error_count = 0;
		status = VProcess_Poweroff;
		lastCheckTime = 0;
	}
	virtual ~KVirtualHostProcess() {
		killProcess(0);
	}
	void isBad(KUpstream *st,BadStage stage)
	{
		if (stage == BadStage_Connect || stage == BadStage_TrySend){
			error_count++;
			if (kgl_current_sec - lastCheckTime > 5) {
				lastCheckTime = kgl_current_sec;
				if(!isProcessActive() || (max_error_count>0 && error_count>=max_error_count)) {
					clean();
					status = VProcess_Close;
					//reset the error_count
					error_count = 0;
					klog(KLOG_ERR,"restart the virtual process error_count=%d\n",error_count);
				}
			}
		}
	}
	void isGood(KUpstream *st)
	{
		//reset the error_count
		error_count = 0;
	}
	virtual kev_result handleRequest(KHttpRequest *rq,KExtendProgram *rd, KAsyncFetchObject *fo);
	KUpstream *connect(KHttpRequest *rq,KExtendProgram *rd, bool &isHalf) {
		lastActive = kgl_current_sec;
		KUpstream *ps = GetPoolSocket(rq);
		if (ps) {
			isHalf = false;
			return ps;
		}
		isHalf = true;
#ifdef KSOCKET_UNIX
		if (!unix_path.empty()) {
			struct sockaddr_un un_addr;
			ksocket_unix_addr(unix_path.c_str(),&un_addr);
			SOCKET fd = ksocket_half_connect((sockaddr_i *)&un_addr,NULL,0);
			if (!ksocket_opened(fd)) {
				return NULL;
			}
			kconnection *cn = kconnection_new(&addr);
			cn->st.fd = fd;
			return new KTcpUpstream(cn);
		}
#endif
		kconnection *cn = kconnection_new(&addr);
		bool result = kconnection_half_connect(cn,NULL,0);		
		if (!result) {
			kconnection_destroy(cn);
			return NULL;
		}
		return new KTcpUpstream(cn);
	}

	virtual KTcpUpstream *poweron(KVirtualHost *vh,KExtendProgram *rd, bool &success) = 0;
	virtual void getProcessInfo(const USER_T &user, const std::string &name,
			std::stringstream &s,int &count) {
	}
	
	virtual bool canDestroy(time_t nowTime)
	{
		if (idleTime>0 && nowTime - lastActive > idleTime) {
			return true;
		}
		return false;
	}
	/*
	ɱָ̣pid=0ɱȫ̡
	true,ʾȫɱ
	false,ʾн(Զ)
	*/
	virtual bool killProcess(int pid) {
		return false;
	}
	/*
	 ʶǷѾ
	 */
	//bool closed;
	int idleTime;
	VProcess_Status status;
	pthread_t workThread;
	KMutex lock;
	int error_count;
	int max_error_count;
	unsigned max_request;
	//u_short port;
	sockaddr_i addr;
	std::list<KVirtualHostProcessQueue *> queue;
	time_t lastCheckTime;
protected:
	time_t lastActive;
	virtual bool isProcessActive()
	{
		return true;
	}
	std::string unix_path;
};
struct VProcessPowerParam
{
	KHttpRequest *rq;
	KVirtualHost *vh;
	KVirtualHostProcess *process;
	KExtendProgram *rd;
	KAsyncFetchObject *fo;
};
void getProcessInfo(const USER_T &user,const std::string &name,KProcess *process,KPoolableSocketContainer *ps,std::stringstream &s);
KTHREAD_FUNCTION VProcessPowerWorker(void *param);
#endif /* KVIRTUALPROCESS_H_ */
