#ifndef _WIN32
#include <sys/select.h>
#include <stddef.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include "ksapi.h"
#include "upstream.h"

static void my_msleep(int msec) {
#if	defined(OSF)
	/* DU don't want to sleep in poll when number of descriptors is 0 */
	usleep(msec * 1000);
#elif	defined(_WIN32)
	Sleep(msec);
#else
	struct timeval tv;
	tv.tv_sec = msec / 1000;
	tv.tv_usec = (msec % 1000) * 1000;
	select(1, NULL, NULL, NULL, &tv);
#endif
}
static int64_t get_post_left(kgl_push_gate *gate, KREQUEST rq)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->get_post_left(ctx->gate, rq);
}
static kev_result read_post(kgl_push_gate *gate, KREQUEST rq)
{
	kgl_async_context *ctx = get_async_context(gate);
	test_context *t = (test_context *)ctx->module;
	if (t->model == test_upload_sleep_forward) {
		t->read_post = true;
		my_msleep(100);		
	}
	return ctx->gate->f->read_post(ctx->gate, rq);
}
static void push_status(kgl_push_gate *gate, KREQUEST rq, int status_code)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->push_status(ctx->gate, rq, status_code);
}
static KGL_RESULT push_header(kgl_push_gate *gate, KREQUEST rq, kgl_header_type attr, const char *val, hlen_t val_len)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->push_header(ctx->gate, rq, attr, val, val_len);
}
static KGL_RESULT push_unknow_header(kgl_push_gate *gate, KREQUEST rq, const char *attr, hlen_t attr_len, const char *val, hlen_t val_len)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->push_unknow_header(ctx->gate, rq, attr, attr_len, val, val_len);
}
static kev_result push_header_finish(kgl_push_gate *gate, KREQUEST rq)
{
	kgl_async_context *ctx = get_async_context(gate);
	test_context *t = (test_context *)ctx->module;
	if (t->read_post) {
		ctx->gate->f->push_unknow_header(ctx->gate, rq, kgl_expand_string("x-testdso"), kgl_expand_string("forward-sleep"));
	}
	return ctx->gate->f->push_header_finish(ctx->gate, rq);
}
static KGL_RESULT push_body(kgl_push_gate *gate, KREQUEST rq, const char *str, int len)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->push_body(ctx->gate, rq, str, len);
}
static kev_result push_body_finish(kgl_push_gate *gate, KREQUEST rq, KGL_RESULT result)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->push_body_finish(ctx->gate, rq, result);
}
static kev_result handle_error(kgl_push_gate *gate, KREQUEST rq, KGL_MSG_TYPE type, int32_t status_code, const char *msg)
{
	kgl_async_context *ctx = get_async_context(gate);
	return ctx->gate->f->push_message(ctx->gate, rq, type, status_code, msg);
}
static void release(kgl_push_gate *gate)
{

}
static kgl_push_gate_function push_gate_function = {
	get_post_left,
	read_post,
	push_status,
	push_header,
	push_unknow_header,
	push_header_finish,
	push_body,
	push_body_finish,
	handle_error,
	release
};
void init_push_gate(kgl_async_context *ctx)
{
	//printf("init_push_gate\n");
	ctx->ug.f = &push_gate_function;
}
